#include "main.h"

// Fuses: clk-div-8 deakivieren, RC-Takt auf 128kHz stellen
// Default: CKSEL3:0=0010    CKDIV8=0
// 128kHz:           0100           1

volatile uint8_t timer_flag;
volatile uint8_t cnt_komparator;



int main (void)
{
    uint8_t test;
	
	
	// fuer Debugging LED
	DDRA|=_BV(PA5);
	DDRB|=_BV(PB2);   // Zaehler-Ausgang

	// LED blinken lassen
	DDRA=_BV(PA5);  // erst mal nur den Portpin fr die LED auf Ausgang schalten
	PORTA|=_BV(PA5); // LED ON;
	PORTB|=_BV(PB2);    // Zahler zaehlen
	for (test=0;test<6;test++)
		{
			PORTA^=_BV(PA5);    // LED ON|OFF
			PORTB^=_BV(PB2);    // Zahler zaehlen
			_delay_ms(250);    // scheint nur fuer 1MHz zu gelten, trotz 8MHz-Eintrag im makefile
			//_delay_ms(250);   //  
			//_delay_ms(250);   //
		}  // LED bleibt am Ende an

	//PORTA|=_BV(PA5); // LED ON
	PORTA&=~_BV(PA5); // LED OFF
	PORTB&=~_BV(PB2);    // Zahler zaehlen

	_delay_ms(1000);
	//sei();		// Interrupts erlauben -> wird selekiv da gemacht wo erforderlich
	init();




	//testroutinen();

	


	while(1)
		{
		work_to_do();
		sleepwait(3000); // 3 Sekunden warten
		}
	


}	





EMPTY_INTERRUPT(ADC_vect); 	// der ADC-interrupt muss aktiviert sein damit der 
				// uC aufwacht




ISR(TIM0_OVF_vect)
{

	
	timer_flag=1;   // zeigt dem Hauptprogramm dass der Sleepmode durch einen
					// Timer 0 Interrupt beendet wurde
					
	// nur fuer die Routine Timertest erforderlich:
	//TCCR0B=0;		// Counter stoppen (Clock source deaktivieren, S.38)
	//PORTA&=~_BV(PA5); // LED OFF (zu testzwecken)

}





ISR(ANA_COMP_vect)
{
	cnt_komparator++;
}







void init(void)
{

	// Timer 0 fuer 250ms programmieren (bei 128kHz Takt)
	// normal mode: zaehlt aufwaerts, TOV flag wird bei 255 gesetzt	
	TIMSK0 = _BV(TOIE0);	// Timer0 overflow interrupt aktivieren
	

	// Komparator konfigurieren, aber keinen Interrupts auslsen lassen
	// ACME = 0 -> AIN1 ist negativer Eingang (S.129)
	ADCSRB &= ~_BV(ACME);  // auch wenns eh default ist
	ACSR |= _BV(ACBG);		// 1,1V Referenz an den pos. Eingang
	
	ACSR &= ~_BV(ACIS1);  // beide Bits =0 -> Komparator reagiert auf Toggeln
	ACSR &= ~_BV(ACIS0);  // auch wenns eh deafult ist
	ACSR |= _BV(ACD);		// Komparator ausschalten, spart Strom im active und idle mode
	
	DIDR0 |= _BV(ADC2D);  // digitalen Input Puffer von AIN1 und AIN0 deaktivieren
	DIDR0 |= _BV(ADC1D);  // um Strom zu sparen wenn analoge Signale anliegen (S.131)



    PRR    = (1 << PRTIM1) | (1 << PRUSI);  // Timer 1 und das USI abschalten,
                                             // den ADC und Timer 0 an lassen

	DDRB|=_BV(PB2);    // Zaehler-Ausgang
	DDRB|=_BV(PB1);    // OPV-Betriebsspannung

	PORTB |= _BV(PB0);		// PB0 Pull up aktivieren (damit der Eingang nicht floatet) 
	PORTA |= 0xDB;			// Port A, auer PA2 und PA5 Pull-Up aktivieren
}







void work_to_do(void)
// diese Funktion wurd alle 3..5 Sekunden aufgerufen und
// macht die eigentliche Arbeit
{
	static uint8_t cnt_adc_meas=222;  // gleich beim ersten mal die Batteriespg. messen
//	static uint8_t cnt_adc_meas=0;  // mit Batteriespg. messen maximal warten
	static uint8_t cnt_totzeit=0;
	static uint8_t bat_status=2;  // Batterie wird bis zur ersten Messung als voll angenommen
	static uint8_t flanken_flag=0;


	// alle 10x die Messung der Bat-Spg aufrufen
	cnt_adc_meas++;
	if (cnt_adc_meas>10)
		{
		cnt_adc_meas=0;
		bat_status=batterytest();  //loest Resets aus!!?!?! 
		}
	if (bat_status==0) // wenn die Zelle ziemlich am Ende ist
		{			// nichts weiter tun um nicht unntig Strom zu verbrauchen
		return;
		}
		

	if (cnt_totzeit>0) // Totzeit dekrementieren
		{
		cnt_totzeit--;
		}



		// Operationsverstaerker einschalten
		PORTB|=_BV(PB1); 

		ACSR &= ~_BV(ACD);		// Komparator einschalten
		sleepwait(50);	// 50ms Wartezeit, so lang braucht die OP-Schaltung
		// u.a. loest der Komparator einen Interrupt aus wenn man den gleich
		// nach dem Einschalten des Komparators aktiviert


		// Komparator konfigurieren, Interrupts auslsen lassen
		ACSR |= _BV(ACI);		// evtl. anstehende Intrruptanfrage loeschen
		ACSR |= _BV(ACIE);		// Komparator-Interrupt aktivieren




		// Timer scharf machen, um eine bestimmte Zeit lang die
		// Komparatorereignisse zu zaehlen
		//PORTA|=_BV(PA5); // LED ON
		cnt_komparator=0;	// in dieser Variable werden die Komparator-Ereignisse gezaehlt
		timer_flag=0;		// in dieser Variable meldet die Timer-ISR dass sie aufgerufen wurde
		sei();				// Interrupts zulassen
		TCNT0 = 128;			// Startwert fuer den Counter
		TCCR0B = _BV(CS02);		// Clock = clk_io/256 (S.83)  -> 500Hz bei 128kHz Takt
				// Sollte 256 ms laufen ... der Interrupt stoppt den Timer

		MCUCR  = (1 << SE);      // CPU sleep mode enable, und Mode auf "Idle" setzen
		while (timer_flag==0)	// solang der Timer nicht abgelaufen ist
			{
			sleep_cpu(); // immer wieder in sleep Modus gehen (nach einen Komparator-Ereignis)
			}
		cli();			// Interrupts sperren
		ACSR &= ~_BV(ACIE);		// Interrupt durch Komparator sperren
		TCCR0B = 0;		// dem Timer den Takt abdrehen
		MCUCR  = 0;      // CPU sleep mode disable, damit der nicht versehentlich ausgeloest wird
		//PORTA&=~_BV(PA5); // LED OFF

		// Komparator deaktivieren
		ACSR &= ~_BV(ACIE);	// Komparator-Interrupt deaktivieren
		ACSR |= _BV(ACD);		// Komparator ausschalten, spart Strom im active und idle mode

		// OPV wieder ausschalten
		PORTB&=~_BV(PB1); // OPV-Betriebsspg.
		
		
		//if (cnt_komparator>8) cnt_komparator=8;
		//sensortest();		// Das gemessene Ergebnis testweise visualisieren

		if (cnt_komparator>6)	// wenn Anzahl Komparator-Ereignisse ueber der Schwelle liegt
			{
			if (flanken_flag==0)	// war das ein neue Ausloesung?
				{
				flanken_flag=1;
				if (cnt_totzeit==0)		// Mindestzeit zwischen zwei Zaehler-
					{					// Ausloesungen schon vestrichen?
					cnt_totzeit=10;	// Totzeit auf 10 Funktionsaufrufe setzen
					zaehler_zwei_schritte(bat_status);	// 2 Toasts zaehlen
					}
				}
			}
		else
			{
			flanken_flag=0;
			}



}




void zaehler_zwei_schritte(uint8_t bat_status)
{	// Zaehlt zwei Toasts weiter

	if (bat_status>0)  // nur wenn die Batterie voll genug ist
		{			// (die Abfrage ist jetzt eigentlich unnoetig, ich habs trotzdem drin gelassen
		PORTA|=_BV(PA5); // LED ON
		PORTB|=_BV(PB2); // Zaehler-Ausgang
		sleepwait(100);
		PORTA&=~_BV(PA5); // LED OFF
		PORTB&=~_BV(PB2); // Zaehler-Ausgang
		sleepwait(200);
		
		if (bat_status==1)  // wenn die Batterie schwach ist
			{				// -> laenger warten zwischen den Zaehlvorgaengen
			sleepwait(1500);
			}
			
		PORTA|=_BV(PA5); // LED ON
		PORTB|=_BV(PB2); // Zaehler-Ausgang
		sleepwait(100);
		PORTA&=~_BV(PA5); // LED OFF
		PORTB&=~_BV(PB2); // Zaehler-Ausgang
		}


}








uint8_t batterytest(void)
// bestimmt die Batteriespannung der Li-Zelle und gibt abhngig davon davon
// ein Flag zurck:
// 0 = Zelle leer, keine Stromfressende aktivitaeten mglich
// 1 = Zelle fast leer
// 2 = Spannung OK, normale Funktionalitaet
// Die Bandgap Referenz wird automatisch durch aktivieren des ADCs mit der Bandgap als
// Eingang aktiviert (Datenblatt Kapitel 7.4.4, S.35)  
{
uint8_t flag;

	sei(); // Interrupts freigeben, damit der ADC die CPU aus dem Schlaf reien kann 
  // enable ADC Noise Reduction Sleep Mode
    MCUCR  = (1 << SE) | (1 << SM0);      // CPU sleep mode enable, und Mode auf "ADC noise reduction" setzen
  // first dummy reading, VRef=Vcc
    //PRR    = (1 << PRTIM1) | (1 << PRTIM0) | (1 << PRUSI);  // die Timer und das USI waehrend des Sleep mode abschalten, den ADC an lassen
    ADMUX  = (1 << MUX5) | (1 << MUX0); //Vcc als Referenz, Messung: die 1,1V Ref-Spg
    ADCSRA = (1 << ADEN) | (1 << ADIE) | (1<< ADPS0); //turn on ADC;ADC completion interrupt;
						// ADC prescaler=2->64kHz bei 128kHz Takt (S.128)
    sleep_cpu(); // dummy ADC reading

    sleep_cpu();  // read battery voltage
    uint16_t bat_volt = ADC;


	if (bat_volt<BAT_SCHWACH)   // Spannung oberhalb BAT_SCHWACH Schwelle? (ADC misst invertiert!)
		{ 
		flag=2;			// -> alles OK
		}
	else
		{
		if (bat_volt<BAT_LEER)  // Spannung oberhalb BAT_LEER Schwelle?
			{
			flag=1;     // -> Batterie ist schwach
			}
		else
			{
			flag=0;     // -> Batterie ist leer
			}
		}



  // disable ADC
    ADCSRA = 0;
    //PRR    = (1 << PRTIM1) | (1 << PRUSI) | (1 << PRADC);
    MCUCR  = 0;      // CPU sleep mode disable, damit der nicht versehentlich ausgeloest wird
	cli(); // Interrupts sperren



return flag;

}






void sleepwait(uint16_t time)
// wartet die angegebene Zeit in Milisekunden (nur gerade), nicht aktiv sondern im Sleepmode
// ber einen Timer gesteuert. Der mu vom globalen Init schon initialisiert sein
{    
uint8_t n;

    PRR    = (1 << PRTIM1) | (1 << PRUSI) | (1 << PRADC);  // den ADC deaktivieren bringt nochmal 4uA
	//timer_flag=0;
	sei();
	TIMSK0 = _BV(TOIE0);	// Timer0 overflow interrupt aktivieren
	TCCR0B = _BV(CS02);		// Clock = clk_io/256 (S.83)  -> 500Hz bei 128kHz Takt
			// Sollte 256 ms laufen ... der Interrupt stoppt den Timer
			// Fuses: 128Khz Clock aktivieren und clk_div_8 deaktivieren

	TCNT0 = 255-((time>>1)&0xff);			// Startwert fuer den Counter
	
	//_delay_ms(100);
	//while (timer_flag==0) {};  // die Variable mu als volatile deklariert sein damit das geht
  
    MCUCR  = (1 << SE);      // CPU sleep mode enable, und Mode auf "Idle" setzen
    //MCUCR  = (1 << SE)|3;      // CPU sleep mode enable, und Mode auf "Standby" setzen
	sleep_cpu();
	
	for(n=(uint16_t)(time>>9);n>0;n--)	// Anzahl der Durchlufe mit 512ms
		{
		//TCNT0 = 0; // nicht benoetigt, Timer laeuft eh ueber
		sleep_cpu(); // warten
		}

	TCCR0B = 0;		// dem Timer den Takt abdrehen
    MCUCR  = 0;     // CPU sleep mode disable, damit der nicht versehentlich ausgeloest wird
	cli();			// Interrupts sperren

    PRR    = (1 << PRTIM1) | (1 << PRUSI); // den ADC wieder aktivieren (Timer0 und USI bleiben immer aus)

return;
}








void stromdetektortest(void)
{	
	// Komparator konfigurieren, aber keinen Interrupts auslsen lassen
	// ACME = 0 -> AIN1 ist negatiber Eingang (S.129)
	ADCSRB &= ~_BV(ACME);  // auch wenns eh default ist
	ACSR |= _BV(ACBG);		// 1,1V Referenz an den pos. Eingang

	ACSR &= ~_BV(ACIS1);  // beide Bits =0 -> Komparator reagiert auf Toggeln
	ACSR &= ~_BV(ACIS0);  // auch wenns eh default ist
	
	//ACSR |= _BV(ACIE);		// Interrupt durch Komparator aktivieren


	
	// alle 0,5s berprfen und LED entsprechend setzen
	while (1)
		{
		if ( ACSR&_BV(ACI) )	// Komparator-Int hat angeschlagen?
			{
			PORTA|=_BV(PA5); // LED ON
			}
		else
			{
			PORTA&=~_BV(PA5); // LED OFF
			}
			
		ACSR|=_BV(ACI);		// Interruptflag zurcksetzen
		
		_delay_ms(250);  // busy waiting
		
		}
	
	
	

}







void adctest(void)
{	
	
	
	// alle 0,5s berprfen und LED entsprechend setzen
	while (1)
		{
		if ( batterytest()==1 )	// Batterietest aufrufen
			{
			PORTA|=_BV(PA5); // LED ON
			}
		else
			{
			PORTA&=~_BV(PA5); // LED OFF
			}
			
		
		_delay_ms(50);  // busy waiting
		
		}
	
	
	

}



void timertest(void)
{
	while(1)
	{
		// Timer 0 fuer 250ms programmieren (bei 128kHz Takt)
		// normal mode: zaehlt aufwaerts, TOV flag wird bei 255 gesetzt
		sei();
		PORTA|=_BV(PA5); // LED ON
		
		TIMSK0 = _BV(TOIE0);	// Timer0 overflow interrupt aktivieren
		TCNT0 = 128;			// Startwert fuer den Counter
		TCCR0B = _BV(CS02);		// Clock = clk_io/256 (S.83)  -> 500Hz bei 128kHz Takt
				// Sollte 256 ms laufen ... der Interrupt stoppt den Timer
				// Fuses: 128Khz Clock aktivieren und clk_div_8 deaktivieren

		_delay_ms(250); 
		_delay_ms(250);   
		_delay_ms(250);   
		_delay_ms(250);   
		_delay_ms(250);   
		_delay_ms(250);   
		_delay_ms(250);   
		//PORTA&=~_BV(PA5); // LED OFF
		_delay_ms(250);   

	}
}



void sleepwaittest(void)
{

	while(1)
		{
		PORTA|=_BV(PA5); // LED ON
		sleepwait(5000);
		PORTA&=~_BV(PA5); // LED OFF
		sleepwait(50);
		}
}





void sensortest(void)
{				// so oft (+1 mal) blinken wie der Komparator aktiviert wurde
	uint8_t n;
	for (n=0;n<cnt_komparator+1;n++)
		{
		PORTA|=_BV(PA5); // LED ON
		sleepwait(50);
		PORTA&=~_BV(PA5); // LED OFF
		sleepwait(100);
		}
}






void testroutinen(void)  // Aufruf einer Testroutine
{
	//sleepwaittest();
	//stromdetektortest();
	adctest();
	//timertest();

	cnt_komparator=0;
	while(1)
		{sensortest();sleepwait(3000);cnt_komparator++;
		}


}

